from typing import Optional

from steamship import Block, SteamshipError, Tag
from steamship.data.tags.tag_constants import ChatTag, DocTag, TagValueKey
from steamship.experimental.easy.tags import get_tag_value_key


class ChatMessage(Block):

    who: Optional[str] = "bot"

    def __init__(self, chat_id: Optional[str] = None, message_id: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)

        # Note: Keep the str() wrappings below since Telegram sends int values.
        # The TelegramTransport will take care of re-int-ifying them.

        if chat_id:
            self.set_chat_id(chat_id)

        if message_id:
            self.set_message_id(message_id)

    @staticmethod
    def from_block(block: Block, chat_id: Optional[str] = None, message_id: Optional[str] = None):
        return ChatMessage(
            chat_id=chat_id,
            message_id=message_id,
            client=block.client,
            id=block.id,
            file_id=block.file_id,
            text=block.text,
            tags=block.tags,
            index_in_file=block.index_in_file,
            mime_type=block.mime_type,
            url=block.url,
            content_url=block.content_url,
            upload_type=block.upload_type,
            upload_bytes=block.upload_bytes,
        )

    def set_chat_id(self, chat_id: str):
        existing = self.get_chat_id()
        if existing is not None:
            if existing == chat_id:
                return  # No action necessary
            else:
                raise SteamshipError(
                    message=f"Block {self.id} already has an existing chat id {existing}. Unable to set to {chat_id}"
                )

        if self.client and self.id:
            tag = Tag.create(
                self.client,
                file_id=self.file_id,
                block_id=self.id,
                kind=DocTag.CHAT,
                name=ChatTag.CHAT_ID,
                value={TagValueKey.STRING_VALUE: str(chat_id)},
            )
        else:
            tag = Tag(
                kind=DocTag.CHAT,
                name=ChatTag.CHAT_ID,
                value={TagValueKey.STRING_VALUE: str(chat_id)},
            )

        self.tags.append(tag)

    def set_message_id(self, message_id: str):
        existing = self.get_message_id()
        if existing is not None:
            if existing == message_id:
                return  # No action necessary
            else:
                raise SteamshipError(
                    message=f"Block {self.id} already has an existing message_id {existing}. Unable to set to {message_id}"
                )

        if self.client and self.id:
            tag = Tag.create(
                self.client,
                file_id=self.file_id,
                block_id=self.id,
                kind=DocTag.CHAT,
                name=ChatTag.MESSAGE_ID,
                value={TagValueKey.STRING_VALUE: str(message_id)},
            )
        else:
            tag = Tag(
                kind=DocTag.CHAT,
                name=ChatTag.MESSAGE_ID,
                value={TagValueKey.STRING_VALUE: str(message_id)},
            )

        self.tags.append(tag)

    def get_chat_id(self) -> str:
        return get_tag_value_key(
            self.tags, TagValueKey.STRING_VALUE, kind=DocTag.CHAT, name=ChatTag.CHAT_ID
        )

    def get_message_id(self) -> str:
        return get_tag_value_key(
            self.tags, TagValueKey.STRING_VALUE, kind=DocTag.CHAT, name=ChatTag.MESSAGE_ID
        )



from typing import Tuple, Union

from steamship import Block, Task, TaskState
from steamship.agents.base import Action, FinishAction
from steamship.agents.context.context import AgentContext
from steamship.agents.planner.base import Planner
from steamship.invocable import PackageService, post


def _is_running(task: Task) -> bool:
    task.refresh()
    return task.state not in [TaskState.succeeded, TaskState.failed]


class AgentService(PackageService):
    agent_context_identifier = (
        "_steamship_agent_contexts"  # probably want to include instance_handle...
    )
    planner: Planner

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    ###############################################
    # Tool selection / execution magic
    ###############################################

    def _next_action(self, context: AgentContext) -> Action:
        return self.planner.plan(context)

    @post("take_action")
    def take_action(self, context: AgentContext) -> Union[Action, FinishAction]:
        action = self._next_action(context)
        if isinstance(action, FinishAction):
            return action

        block_or_task = action.tool.run(action.tool_input, action.context)
        if isinstance(block_or_task, Task):
            context.in_progress.append((action, block_or_task))
            # self.upsert_context(context)

            for func in context.emit_funcs:
                func([Block(text=f"Status: Running {action}")], context.metadata)
            self.invoke_later(
                method="_steamship/run",
                arguments={"context": context},
                wait_on_tasks=[block_or_task],
            )
        else:
            output_blocks = action.tool.run(action.tool_input, context)
            action.tool_output = output_blocks
            context.completed_steps.append(action)
            # self.upsert_context(context)
            return action

    @post("_steamship/run")
    def run_agent(self, context: AgentContext):
        # first thing we must do is determine if we are still waiting on any pending executions
        # if so, reschedule run_agent for later.
        last_known_pending_tasks = context.in_progress
        completed_bindings_and_tasks = [
            Tuple[tb, t] for tb, t in last_known_pending_tasks if not _is_running(t)
        ]
        for action, task in completed_bindings_and_tasks:
            task.refresh()
            action.tool_output = task.output.blocks
            context.completed_steps.append(action)
        # self.upsert_context(context)
        running_tasks = [t for _, t in last_known_pending_tasks if _is_running(t)]
        if len(running_tasks) > 0:
            # todo: do we need to do anything with this pending task?
            self.invoke_later(
                method="_steamship/run", arguments={"context": context}, wait_on_tasks=running_tasks
            )
            return

        # at this stage in run_agent, we have no more running tasks. we have 0 or more completed steps.
        # we must now select the next step
        action = self.take_action(context)
        while not isinstance(action, FinishAction):
            action = self.take_action(context)

        for func in context.emit_funcs:
            func(action.output, context.metadata)
        # self.unload_context(context_id=context.id)



from abc import ABC, abstractmethod
from typing import List, Union

from pydantic import BaseModel

from steamship.agents.base import LLM, Action, AgentContext, BaseTool, FinishAction
from steamship.agents.parsers.base import OutputParser


class Planner(BaseModel, ABC):
    tools: List[BaseTool]

    @abstractmethod
    def plan(self, context: AgentContext) -> Union[Action, FinishAction]:
        pass


class LLMPlanner(Planner):
    llm: LLM
    # input_preparer: InputPreparer
    output_parser: OutputParser

    @abstractmethod
    def plan(self, context: AgentContext) -> Action:
        # sketch...
        # prompt = PROMPT.format(input_preparer.prepare(context))
        # generation = llm.generate(prompt)
        # tool_name, inputs = output_parser.parse(generation)
        # return (tools[tool_name], inputs)
        pass



"""Tests that base classes can register routes, but they're overridden by subclasses.

This tests a fix (in the same PR) to a behavior in which each new Subclass of Invocable would register its route
table afresh, overwriting whatever routes were registered by an ancestor.
"""

from steamship.invocable import Invocable, InvocableRequest, Invocation, post
from steamship.utils.url import Verb


class L1Invocable(Invocable):
    @post("foo")
    def foo(self) -> str:
        return "l1_foo"

    @post("bar")
    def bar(self) -> str:
        return "l1_bar"

    @post("baz")
    def baz(self) -> str:
        return "l1_baz"


class L2Invocable(L1Invocable):
    @post("bar")
    def bar(self) -> str:
        return "l2_bar"

    @post("baz")
    def baz(self) -> str:
        return "l2_baz"


class L2Invocable2(L1Invocable):
    @post("bar")
    def bar2(self) -> str:
        return "l22_bar"

    @post("baz")
    def baz2(self) -> str:
        return "l22_baz"


class L3Invocable(L2Invocable):
    @post("baz")
    def baz(self) -> str:
        return "l3_baz"


class L3Invocable2(L2Invocable2):
    @post("baz")
    def baz3(self) -> str:
        return "l32_baz"


def invoke(o: Invocable, path: str):
    req = InvocableRequest(invocation=Invocation(http_verb="POST", invocation_path=path))
    return o(req)


def test_l1_routes():
    """Tests that we can inspect the L1 routes"""
    l1 = L1Invocable()
    assert l1._method_mappings[Verb.POST]["/foo"] == "foo"
    assert l1._method_mappings[Verb.POST]["/bar"] == "bar"
    assert l1._method_mappings[Verb.POST]["/baz"] == "baz"
    assert invoke(l1, "foo") == "l1_foo"
    assert invoke(l1, "bar") == "l1_bar"
    assert invoke(l1, "baz") == "l1_baz"


def test_l2_routes():
    """Tests that we can inspect the L1 routes"""
    l2 = L2Invocable()
    assert l2._method_mappings[Verb.POST]["/foo"] == "foo"
    assert l2._method_mappings[Verb.POST]["/bar"] == "bar"
    assert l2._method_mappings[Verb.POST]["/baz"] == "baz"
    assert invoke(l2, "foo") == "l1_foo"
    assert invoke(l2, "bar") == "l2_bar"
    assert invoke(l2, "baz") == "l2_baz"


def test_l3_routes():
    """Tests that we can inspect the L1 routes"""
    l3 = L3Invocable()
    assert l3._method_mappings[Verb.POST]["/foo"] == "foo"
    assert l3._method_mappings[Verb.POST]["/bar"] == "bar"
    assert l3._method_mappings[Verb.POST]["/baz"] == "baz"
    assert invoke(l3, "foo") == "l1_foo"
    assert invoke(l3, "bar") == "l2_bar"
    assert invoke(l3, "baz") == "l3_baz"


def test_l22_routes():
    """Tests that we can inspect the L1 routes"""
    l22 = L2Invocable2()
    assert l22._method_mappings[Verb.POST]["/foo"] == "foo"
    assert l22._method_mappings[Verb.POST]["/bar"] == "bar2"
    assert l22._method_mappings[Verb.POST]["/baz"] == "baz2"
    assert invoke(l22, "foo") == "l1_foo"
    assert invoke(l22, "bar") == "l22_bar"
    assert invoke(l22, "baz") == "l22_baz"


def test_l32_routes():
    """Tests that we can inspect the L1 routes"""
    l32 = L3Invocable2()
    assert l32._method_mappings[Verb.POST]["/foo"] == "foo"
    assert l32._method_mappings[Verb.POST]["/bar"] == "bar2"
    assert l32._method_mappings[Verb.POST]["/baz"] == "baz3"
    assert invoke(l32, "foo") == "l1_foo"
    assert invoke(l32, "bar") == "l22_bar"
    assert invoke(l32, "baz") == "l32_baz"



"""Tool for generating images."""
from steamship.agents.tool import AudioBlockifierTool
from steamship.utils.repl import ToolREPL


class AssemblySpeechToTextTool(AudioBlockifierTool):
    """Tool to generate text from audio."""

    name: str = "AssemblySpeechToTextTool"
    human_description: str = "Generates text from spoken audio."
    ai_description: str = (
        "Used to generate text from spoken audio. Only use if the user has asked directly for a "
        "an text version of an audio file. When using this tool, the input should be the audio file. "
        "The output is the text"
    )
    blockifier_plugin_handle: str = "s2t-blockifier-assembly"


if __name__ == "__main__":
    ToolREPL(AssemblySpeechToTextTool()).run()
    # Try on: https://anchor.fm/s/e1369b4c/podcast/play/70381739/https%3A%2F%2Fd3ctxlq1ktw2nl.cloudfront.net%2Fstaging%2F2023-4-14%2Fda96d64d-3c27-3e84-eb63-d5bc55eaf52e.mp3



import re
from typing import Dict, Optional, Union

from steamship import Block
from steamship.agents.base import Action, AgentContext, BaseTool, FinishAction
from steamship.agents.parsers.base import OutputParser
from steamship.experimental.transports.chat import ChatMessage


class LLMToolOutputParser(OutputParser):
    tools_lookup_dict: Optional[Dict[str, BaseTool]] = None

    def __init__(self, **kwargs):
        tools_lookup_dict = {tool.name: tool for tool in kwargs.pop("tools", None)}
        super().__init__(tools_lookup_dict=tools_lookup_dict, **kwargs)

    def parse(self, text: str, context: AgentContext) -> Union[Action, FinishAction]:
        if "AI:" in text:
            return FinishAction(
                output=[ChatMessage(text=text.split("AI:")[-1].strip())], context=context
            )

        regex = r"Action: (.*?)[\n]*Action Input: (.*)"
        match = re.search(regex, text)
        if not match:
            raise RuntimeError(f"Could not parse LLM output: `{text}`")
        action = match.group(1)
        action_input = match.group(2).strip()
        return Action(
            tool=self.tools_lookup_dict[action.strip()],
            tool_input=[Block(text=action_input)],
            context=context,
        )



from abc import ABC, abstractmethod
from typing import Union

from pydantic import BaseModel

from steamship.agents.base import Action, AgentContext, FinishAction


class InputParser(BaseModel, ABC):
    # This takes multi-media and converts into a textual form useful for prompting
    # for further text generation.
    #
    # example:
    #  - input: AgentContext(completed_steps: [Step(dall-e, prompt, output_block)])
    #  - output: "Action: Dall-E Tool
    #             ActionInput: prompt
    #             ActionOutput: Block(UUID)"
    @abstractmethod
    def prepare(self, context: AgentContext) -> str:
        pass


class OutputParser(BaseModel, ABC):
    # Example image-based workflow:
    # (1) user request: generate an image of a row-house in a city street.
    # (2) llm planner: dalle("row-house in a city street")
    # (3) dall-e tool: output=block(02342342342)
    # (4) package response: <image>
    # (5) user request: make the door of the house red/
    # (6) llm planner: pix-to-pix(prompt="make door red", block=02423452345)
    # (7) pix-2-pix tool: output=block(992341234)
    #
    # In this flow, the LLM planner needs to tell the pix-2-pix tool which image to start with, so it has to know how
    # to represent that in text.
    #
    # example:
    #  - input: "Action: Pix-2-Pix
    #            ActionInput: Block(UUID)"
    #  - output: ("pix-to-pix", [Block(UUID)])
    @abstractmethod
    def parse(self, text: str, context: AgentContext) -> Union[Action, FinishAction]:
        pass



import json
import logging
from os import path
from typing import Any, Callable, Dict, Optional

import click

from steamship import Steamship, SteamshipError
from steamship.cli.ship_spinner import ship_spinner
from steamship.data.manifest import DeployableType, Manifest
from steamship.utils.metadata import hash_dict
from steamship.utils.utils import create_instance_handle


class LoggingDisabled:
    """Context manager that turns off logging within context."""

    def __enter__(self):
        logging.disable(logging.CRITICAL)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)


@click.command(name="use")
@click.option(
    "--workspace",
    "-w",
    required=False,
    type=str,
    help="Workspace handle. The new instance will be created in this workspace. If not specified, "
    "the default workspace will be used.",
)
@click.option(
    "--instance",
    "-i",
    type=str,
    required=False,
    help="Instance handle. Used to name the instance (if it needs to be created). If not specified, "
    "a generated instance name will be used.",
)
@click.option(
    "--config",
    "-c",
    type=str,
    required=False,
    help="Instance configuration. May be inline JSON or a path to a file. If not specified, "
    "an empty configuration dictionary will be passed to the instance.",
)
def create_instance(
    workspace: Optional[str] = None, instance: Optional[str] = None, config: Optional[str] = None
):
    """Create an instance of your package/plugin for use.

    Must be run from a directory containing a Steamship manifest.
    """

    with LoggingDisabled():  # keep CLI output as concise as possible
        return _create_instance(workspace, instance, config)


def _create_instance(
    workspace: Optional[str] = None, instance: Optional[str] = None, config: Optional[str] = None
):
    manifest = None
    if path.exists("steamship.json"):
        manifest = Manifest.load_manifest()
    else:
        click.secho("No manifest found for instance creation.", fg="red")
        click.secho("Please try again from a directory with a package or plugin manifest.")
        click.get_current_context().abort()

    if manifest is None:
        click.secho("Steamship manifest failed to load.", fg="red")
        click.get_current_context().abort()

    invocable_config = config_dict(config)

    if workspace is None or len(workspace) == 0:
        ws_hash = hash_dict(
            {
                **invocable_config,
                "version": manifest.version,
                "invocable": manifest.handle,
            }
        )
        workspace = f"space-{ws_hash}"

    client = None
    try:
        client = Steamship(workspace=workspace)
    except SteamshipError as e:
        click.secho(e.message, fg="red")
        click.get_current_context().abort()

    instance_fn = client.use
    if manifest.type == DeployableType.PLUGIN:
        instance_fn = client.use_plugin

    if instance is None:
        instance = create_instance_handle(
            invocable_handle=manifest.handle,
            version_handle=manifest.version,
            invocable_config=invocable_config,
        )

    try:
        _call_fn(client, manifest, instance, invocable_config, instance_fn)

    except SteamshipError as e:
        click.secho(f"\nFailed to create instance: {e.message}", fg="red")
        click.get_current_context().abort()


def _call_fn(
    client: Steamship,
    manifest: Manifest,
    instance: str,
    config: Optional[Dict[str, Any]],
    instance_fn: Callable,
):
    click.echo("Using values:\n- Workspace: ", nl=False)
    click.secho(f"{client.get_workspace().handle}", fg="green")
    click.echo(f"- {manifest.type.title()}: ", nl=False)
    click.secho(f"{manifest.handle}", fg="green")
    click.echo("- Version: ", nl=False)
    click.secho(f"{manifest.version}", fg="green")
    click.echo("- Instance Handle: ", nl=False)
    click.secho(f"{instance}", fg="green")
    click.echo("- Configuration: ", nl=False)
    click.secho(f"{json.dumps(config)}", fg="green")

    click.echo("Creating new (or fetching existing) instance: ", nl=False)
    with ship_spinner():
        instance = instance_fn(
            manifest.handle,
            instance_handle=instance,
            version=manifest.version,
            config=config,
            wait_for_init=True,
        )
        if instance:
            click.secho("\nSuccess!", fg="green")
            if manifest.type == DeployableType.PACKAGE:
                click.echo(f"Instance URL: {instance.invocation_url}")
            return
        else:
            raise SteamshipError("instance creation unexpectedly returned empty instance.")


def config_dict(config: Optional[str]) -> Dict[str, Any]:
    return_dict = {}
    if config:
        try:
            json_config = json.loads(config)
            if isinstance(json_config, dict):
                return_dict = json_config
            else:
                click.secho("Could not parse configuration.", fg="red")
                click.get_current_context().abort()
        except Exception:
            try:
                with open(config) as f:
                    file_json = json.load(f)
                    if isinstance(file_json, dict):
                        return_dict = file_json
                    else:
                        click.secho("Could not parse configuration.", fg="red")
                        click.get_current_context().abort()
            except Exception as e:
                click.secho(
                    f"Unknown configuration. Could not parse it as a file or JSON string: {e}",
                    fg="red",
                )
                click.get_current_context().abort()
    return return_dict



from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel

from steamship import Block, Task
from steamship.agents.context.context import AgentContext


class Action:
    pass  # Circular dependency


ToolInputs = List[Block]
ToolOutputs = List[Block]
AgentSteps = List[Action]
Metadata = Dict[str, Any]
EmitFunc = Callable[[ToolOutputs, Metadata], None]


class BaseTool(BaseModel, ABC):
    # Working thinking: we don't yet have formalization about whether
    # this is a class-level name, isntance-level name, or
    # instance+context-level name.
    # thought(doug): this should be the planner-facing name (LLM-friendly?)
    name: str

    # Advice, but not hard-enforced:
    # This contains the description, inputs, and outputs.
    ai_description: str
    human_description: str  # Human readable string for logging

    @abstractmethod
    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        raise NotImplementedError()

    # This gets called later if you return Task[Any] from run
    def post_process(self, async_task: Task, context: AgentContext) -> List[Block]:
        # nice helpers for making lists of blocks
        pass


class Action(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    tool: BaseTool  # Tools are retrieved via their name
    tool_input: List[Block]  # Tools always get strings as input
    context: AgentContext
    tool_output: Optional[ToolOutputs] = []


class FinishAction(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    output: Any  # Output can be anything as long as it's JSON serializable
    context: AgentContext


class LLM(BaseModel, ABC):
    @abstractmethod
    def complete(self, prompt: str, stop: str) -> List[Block]:
        pass



from typing import List

from steamship import Block, PluginInstance, Steamship
from steamship.agents.base import LLM

PLUGIN_HANDLE = "gpt-4"


class OpenAI(LLM):
    generator: PluginInstance
    client: Steamship

    def __init__(self, client, model_name: str = "gpt-3.5-turbo", *args, **kwargs):
        """Create a new instance.

        Valid model names are:
         - gpt-4
         - gpt-3.5-turbo
        """
        client = client
        generator = client.use_plugin(PLUGIN_HANDLE, config={"model": model_name})
        super().__init__(client=client, generator=generator, *args, **kwargs)

    def complete(self, prompt: str, stop: str) -> List[Block]:
        action_task = self.generator.generate(text=prompt, options={"stop": stop})
        action_task.wait()
        return action_task.output.blocks



import json
from abc import abstractmethod
from typing import Any, List, Optional, Union

import requests

from steamship import Block, File, MimeTypes, Task
from steamship.agents.base import BaseTool
from steamship.agents.context.context import AgentContext


class Tool(BaseTool):
    # Working thinking: we don't yet have formalization about whether
    # this is a class-level name, isntance-level name, or
    # instance+context-level name.
    # thought(doug): this should be the planner-facing name (LLM-friendly?)
    name: str

    # Advice, but not hard-enforced:
    # This contains the description, inputs, and outputs.
    ai_description: str
    human_description: str  # Human readable string for logging

    @abstractmethod
    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        raise NotImplementedError()

    # This gets called later if you return Task[Any] from run
    def post_process(self, async_task: Task, context: AgentContext) -> List[Block]:
        # nice helpers for making lists of blocks
        pass


class GeneratorTool(Tool):
    """
    A base class for tools that wrap Steamship Generator plugins. Subclass this and implement
    the `accept_output_block` method.
    """

    generator_plugin_handle: str
    generator_plugin_instance_handle: Optional[str] = None
    generator_plugin_config: dict = {}
    merge_blocks: bool = False

    @abstractmethod
    def accept_output_block(self, block: Block) -> bool:
        raise NotImplementedError()

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        generator = context.client.use_plugin(
            plugin_handle=self.generator_plugin_handle,
            instance_handle=self.generator_plugin_instance_handle,
            config=self.generator_plugin_config,
        )

        tasks = []

        if self.merge_blocks:
            block = tool_input[0]
            for extra_block in tool_input[1:]:
                block.text = f"{block.text}\n\n{extra_block.text}"
            tool_input = [block]

        for block in tool_input:
            if isinstance(block, str):
                print(block)
            if not block.is_text():
                continue

            prompt = block.text
            task = generator.generate(text=prompt, append_output_to_file=True)
            tasks.append(task)

        # TODO / REMOVE Synchronous execution is a temporary simplification while we merge code.
        output_blocks = []
        # NB: In an async framework, we need to contend with the fact that we have some Monad-style trickery
        # to contend with. Namely that the below code has taken us from List[Block] as the universal type to
        # List[List[Block]] as the async result that we will have to process.
        #
        # It's unclear to me (ted) if this is something we leave as an exercise for implementors or build in
        # as universally handled.
        for task in tasks:
            task.wait()
            task_blocks = self.post_process(task, context)
            for block in task_blocks:
                output_blocks.append(block)
        # END TODO / REMOVE

        return output_blocks

    def post_process(self, task: Task, context: AgentContext) -> List[Block]:
        """In this case, the Generator returns a GeneratorResponse that has a .blocks method on it"""
        try:
            return task.output.blocks
        except BaseException:
            return json.loads(task.output).get("blocks")


class ImageGeneratorTool(GeneratorTool):
    """
    A base class for tools that wrap Steamship Image Generator plugins.
    """

    def accept_output_block(self, block: Block) -> bool:
        return block.is_image()


class AudioGeneratorTool(GeneratorTool):
    """
    A base class for tools that wrap Steamship Audio Generator plugins.
    """

    def accept_output_block(self, block: Block) -> bool:
        return block.is_audio()


class ScrapeAndBlockifyTool(Tool):
    """
    A base class for tools that wrap Steamship Blockifier plugin which transforms bytes to a set of blocks.
    """

    blockifier_plugin_handle: str
    blockifier_plugin_instance_handle: Optional[str] = None
    blockifier_plugin_config: dict = {}

    @abstractmethod
    def should_blockify(self, block: Block) -> bool:
        raise NotImplementedError()

    def get_mime_type(self):
        return None

    def _scrape(self, url: str, context: AgentContext) -> File:
        response = requests.get(url)
        file = File.create(context.client, content=response.content, mime_type=self.get_mime_type())
        return file

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        tasks = []

        blockifier = context.client.use_plugin(
            plugin_handle=self.blockifier_plugin_handle,
            instance_handle=self.blockifier_plugin_instance_handle,
            config=self.blockifier_plugin_config,
        )

        for input_block in tool_input:
            if not input_block.is_text():
                continue

            url = input_block.text
            file = self._scrape(url, context)
            task = file.blockify(blockifier.handle)
            tasks.append(task)

        # TODO / REMOVE Synchronous execution is a temporary simplification while we merge code.
        output_blocks = []
        # NB: In an async framework, we need to contend with the fact that we have some Monad-style trickery
        # to contend with. Namely that the below code has taken us from List[Block] as the universal type to
        # List[List[Block]] as the async result that we will have to process.
        #
        # It's unclear to me (ted) if this is something we leave as an exercise for implementors or build in
        # as universally handled.
        for task in tasks:
            task.wait()
            task_blocks = self.post_process(task, context)
            for block in task_blocks:
                output_blocks.append(block)
        # END TODO / REMOVE

        return output_blocks

    def post_process(self, task: Task, context: AgentContext) -> List[Block]:
        """In this case, the Blockifier returns a BlockAndTagResponse that has a .file.blocks method on it"""
        try:
            return task.output.file.blocks
        except BaseException:
            return json.loads(task.output).get("file", {}).get("blocks")


class ImageBlockifierTool(ScrapeAndBlockifyTool):
    """
    A base class for tools that wrap Steamship Image Blockifier plugins.
    """

    def get_mime_type(self):
        return MimeTypes.PNG

    def should_blockify(self, block: Block) -> bool:
        return block.is_image()


class AudioBlockifierTool(ScrapeAndBlockifyTool):
    """
    A base class for tools that wrap Steamship Audio Blockifier plugins.
    """

    def get_mime_type(self):
        return MimeTypes.MP3

    def should_blockify(self, block: Block) -> bool:
        return block.is_audio()



from typing import List, Optional

from steamship import Block
from steamship.agents.base import Metadata
from steamship.agents.context.context import AgentContext
from steamship.agents.llm.openai import OpenAI
from steamship.agents.planner.react import ReACTPlanner
from steamship.agents.service.agent_service import AgentService
from steamship.agents.tools.image_generation.dalle import DalleTool
from steamship.agents.tools.search.search import SearchTool
from steamship.experimental.transports.chat import ChatMessage

# AgentService is a PackageService.
from steamship.utils.repl import AgentREPL


class MyAssistant(AgentService):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.planner = ReACTPlanner(
            tools=[
                SearchTool(),
                DalleTool(),
            ],
            llm=OpenAI(self.client),
        )

    def create_response(self, context: AgentContext) -> Optional[List[ChatMessage]]:
        # todo: do we need more here to allow for user-specific contexts?
        # todo: how to deal with overlapping requests (do this... and do this... and do this...)

        if len(context.emit_funcs) == 0:
            context.emit_funcs.append(self._send_message_agent)

        # pull up User<->Agent chat history, and append latest Human Input
        # this is distinct from any sort of history related to agent execution
        # chat_file = ChatFile.get(...)
        # chat_file.append_user_block(text=incoming_message.text)
        # current_context.chat_history = chat_file

        self.run_agent(context)

        # should we return any message to the user to indicate that a response?
        # maybe: "Working on it..." or "Received: {prompt}..."
        return []

    def _send_message_agent(self, blocks: List[Block], meta: Metadata):
        # should this be directly-referenced, or should this be an invoke() endpoint, with a value passed
        # in?
        chat_id = meta.get("chat_id")
        message_id = meta.get("message_id")
        messages = [
            ChatMessage.from_block(block=b, chat_id=chat_id, message_id=message_id) for b in blocks
        ]

        # Here is where we would update the ChatFile...
        # chat_file = ChatFile.get(...)
        # chat_file.append_system_blocks(blocks)

        print(f"\n\nTELEGRAM SENDING MESSAGES:\n{messages}")
        # self.telegram_transport.send(messages)


if __name__ == "__main__":
    AgentREPL(MyAssistant).run()



from abc import ABC
from typing import Any, Callable, Dict, List, Optional, Tuple

from pydantic.main import BaseModel

from steamship import Block, PluginInstance, Steamship, Tag, Task
from steamship.agents.context.chathistory import ChatHistory


class BaseTool(BaseModel, ABC):
    pass


class ToolBinding(BaseModel):
    tool: BaseTool
    inputs: List[Block]


ToolInputs = List[Block]
ToolOutputs = List[Block]
AgentStep = Tuple[ToolBinding, ToolOutputs]
AgentSteps = List[AgentStep]
Metadata = Dict[str, Any]
EmitFunc = Callable[[ToolOutputs, Metadata], None]


class AgentContext:
    # AgentContext is passed as a parameter for any `run_agent()` call on the `AgentService`
    # as well as any `Tool.run()` call. It provides access to arbitrary metadata about the
    # context of any agent invocation, as well as anything relevant to the `AgentService`
    # instance etc.
    #
    # It also provides access to two-distinct types of "history" or "memory":
    # (1) user <-> package interactions
    # (2) agent and tool interactions
    #
    # It is meant to be persisted (via a KeyValueStore or otherwise) to allow for
    # restarts, interruptions, etc.

    # A steamship-assigned ID for this context object
    @property
    def id(self) -> str:
        return self.chat_history.file.id

    metadata: Metadata = {}

    client: Steamship

    # User<->Package chat history (NOT Agent<-->Tool history)
    # NB: This is distinct from any sort of history related to agent execution
    # Primarily used to ferry prompts into agent executions AND
    # to allow for some sort of referential lookup of context as input to the agent
    chat_history: ChatHistory

    # Provides the overall goal of the agent
    @property
    def initial_prompt(self) -> Optional[Block]:
        return self.chat_history.initial_system_prompt

    # instead of saving the execution history via File as text, this chooses to represent
    # state as objects. this may not be what we want. This could also be:
    # agent_history: ChatFile (with tags on blocks to represent state of Tool execution)
    # i need to think about that a bit more.
    #
    # for each of these bits, we probably need to save only IDs, and rehydrate properly;
    # likely requires an override of `dict()` and `parse_obj()`
    completed_steps: AgentSteps = []

    # This supports parallel tool execution. But I think that should
    # be a future iteration, as it complicates things quite a bit.
    #
    # as above this likely could be "discovered" from `agent_history: ChatFile` and tags
    # dynamically, and be more "Steamship-native". Perhaps in a second pass.
    in_progress: List[
        Tuple[Any, Task]
    ] = []  # todo: should this be a map from task_id -> ToolBinding?
    # TODO: Any here should be Action, but I can't make it un-circular

    # this, I think(?), is not serializable, and must be set in some sort of context init bit
    # of whatever is doing the work.
    # in the future, this could be a set of callbacks, more broken out (onError, onComplete, ...)
    emit_funcs: List[EmitFunc] = []

    def get_llm(self) -> PluginInstance:
        # This may be something we wish to eventually provide application-level settings for.
        # E.g. the agent has a set_default_llm method that is available and supported in the UI.
        return self.client.use_plugin("gpt-4", config={"model": "gpt-3.5-turbo"})

    @staticmethod
    def get_or_create(
        client: Steamship,
        context_keys: Dict[str, str],
        tags: List[Tag] = None,
    ) -> "AgentContext":
        history = ChatHistory.get_or_create(client, context_keys, tags)
        context = AgentContext()
        context.chat_history = history
        context.client = client
        return context



import abc
import contextlib
import logging
import uuid
from abc import ABC
from typing import List, Optional, Type, cast

from steamship import Block, Steamship, Task
from steamship.agents.base import AgentContext, BaseTool
from steamship.agents.service.agent_service import AgentService
from steamship.data.workspace import SignedUrl, Workspace
from steamship.utils.signed_urls import upload_to_signed_url


class SteamshipREPL(ABC):
    """Base class for building REPLs that facilitate running Steamship code in the IDE."""

    client: Steamship

    def _make_public_url(self, block):
        filepath = str(uuid.uuid4())
        signed_url = (
            self.client.get_workspace()
            .create_signed_url(
                SignedUrl.Request(
                    bucket=SignedUrl.Bucket.PLUGIN_DATA,
                    filepath=filepath,
                    operation=SignedUrl.Operation.WRITE,
                )
            )
            .signed_url
        )
        logging.info(f"Got signed url for uploading block content: {signed_url}")
        read_signed_url = (
            self.client.get_workspace()
            .create_signed_url(
                SignedUrl.Request(
                    bucket=SignedUrl.Bucket.PLUGIN_DATA,
                    filepath=filepath,
                    operation=SignedUrl.Operation.READ,
                )
            )
            .signed_url
        )
        upload_to_signed_url(signed_url, block.raw())
        return read_signed_url

    def print_blocks(self, blocks: List[Block]):
        """Print a list of blocks to console."""
        for block in blocks:
            if isinstance(block, dict):
                block = Block.parse_obj(block)
            if block.is_text():
                print(block.text)
            elif block.url:
                print(block.url)
            elif block.content_url:
                print(block.content_url)
            else:
                url = self._make_public_url(block)
                print(url)

    @contextlib.contextmanager
    def temporary_workspace(self) -> Steamship:
        workspace = Workspace.create(client=self.client)
        temp_client = Steamship(workspace=workspace.handle)
        yield temp_client
        workspace.delete()

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError()


class ToolREPL(SteamshipREPL):
    tool: BaseTool
    client = Steamship

    def __init__(self, tool: BaseTool, client: Optional[Steamship] = None):
        super().__init__()
        self.tool = tool
        self.client = client or Steamship()

    def run_with_client(self, client: Workspace):
        try:
            from termcolor import colored  # noqa: F401
        except ImportError:

            def colored(message: str, color: str):
                print(message)

        context = AgentContext()
        context.client = client

        print(f"Starting REPL for Tool {self.tool.name}...")
        print("If you make code changes, restart this REPL. Press CTRL+C to exit at any time.\n")

        while True:
            input_text = input(colored("Input: ", "blue"))  # noqa: F821
            input_block = Block(text=input_text)
            output = self.tool.run([input_block], context=context)
            if isinstance(output, Task):
                # TODO: Iterate on task support.
                print(f"Task: {output.task_id}")
            else:
                blocks = cast(List[Block], output)
                self.print_blocks(blocks)

    def run(self):
        with self.temporary_workspace() as client:
            self.run_with_client(client)


class AgentREPL(SteamshipREPL):
    agent_class: Type[AgentService]
    client = Steamship

    def __init__(self, agent_class: Type[AgentService], client: Optional[Steamship] = None):
        super().__init__()
        self.agent_class = agent_class
        self.client = client or Steamship()

    def run_with_client(self, client: Steamship):
        try:
            from termcolor import colored  # noqa: F401
        except ImportError:

            def colored(message: str, color: str):
                print(message)

        chat_id = uuid.uuid4().hex

        print("Starting REPL for Agent...")
        print(f"Chat ID: {chat_id}")
        print("If you make code changes, restart this REPL. Press CTRL+C to exit at any time.\n")

        agent = self.agent_class(client=client)

        while True:
            input_text = input(colored("Input: ", "blue"))  # noqa: F821
            message_id = uuid.uuid4().hex

            context = AgentContext.get_or_create(
                client,
                context_keys={
                    "chat_id": chat_id
                    # No message id here; we don't want a new context per message
                },
            )
            message = context.chat_history.append_user_message(
                text=input_text
            )  # Should this take a Block, instead of creating a block?
            message.set_message_id(message_id)
            response: Optional[List[Block]] = agent.create_response(context)
            self.print_blocks(response)

    def run(self):
        with self.temporary_workspace() as client:
            self.run_with_client(client)
